/*
 * Copyright 2001-2005 SageTV, LLC. All rights reserved.
 */
// Standard Microsoft include
#include "stdafx.h"

#include <jni.h>

// Include MS headers for using Managed .NET code
#using <mscorlib.dll>
#include <vcclr.h>

// Include required for Image operations used in this file
#using <System.Drawing.dll>

#include "ManagedToJava.h"

// These struct definitions resolve a TypeLoadException in the managed code that would otherwise
// occur at runtime.
struct _jmethodID {};
struct _jfieldID {};

jstring MgdStringToJString(JNIEnv* env, System::String *mgdString)
{
	if (!mgdString) return NULL;

	// Use the Marshal class to get an unmanaged pointer to an array of unicode characters that
	// represent this String.
	const wchar_t* chars = (const wchar_t*)(Marshal::StringToHGlobalUni(mgdString)).ToPointer();

	// Create the Java string that we'll return from the unicode character array
	jstring jrv = env->NewString(chars, (jsize)wcslen(chars));

	// Free the marshalled unicode character array since we don't need it anymore
	Marshal::FreeHGlobal(IntPtr((void*)chars));

	// Return the Java string
	return jrv;
}

jobject MgdImageToBufferedImage(JNIEnv* env, Bitmap* mgdBitmap)
{
	// We want to create a java.awt.image.BufferedImage object the same size as the managed
	// image object. We'll then have to get unmanaged pointers to the image data from both objects
	// and copy the memory between them.

	// Get the dimensions of the managed image
	int imageWidth = mgdBitmap->get_Width();
	int imageHeight = mgdBitmap->get_Height();

	// Get the class & method identifiers to the constructor for 
	// java.awt.image.BufferedImage(int width, int height, int format)
	jclass buffImageClass = env->FindClass("java/awt/image/BufferedImage");
	jmethodID buffImageConstruct = env->GetMethodID(buffImageClass, "<init>", "(III)V");
	// Create the java.awt.image.BufferedImage that we'll be returning,
	// this is just like:
	// new java.awt.image.BufferedImage(imageWidth, imageHeight, java.awt.image.BufferedImage.TYPE_INT_ARGB);
	jobject rv = env->NewObject(buffImageClass, buffImageConstruct, imageWidth, imageHeight, 
		2); // 2 is TYPE_INT_ARGB from java.awt.image.BufferedImage

	// Create a Rectangle object which specifies that we want to get all of the image data from the managed image
	System::Drawing::Rectangle fullImageRect = System::Drawing::Rectangle(0, 0, imageWidth, imageHeight);

	// Call the method LockBits on the managed image object which will give us back a BitmapData* which
	// we can then use to get an unmanaged pointer to the managed image data. Use an image format that
	// matches the one for the Java BufferedImage object we created above.
	BitmapData* mgdBitmapData = mgdBitmap->LockBits(fullImageRect,
		ImageLockMode::ReadOnly, Imaging::PixelFormat::Format32bppArgb);

	// Now get access to the int array which holds the data for this Java image. This is done on a
	// java.awt.image.BufferedImage named buffImage by calling:
	// ((java.awt.image.DataBufferInt) buffImage.getRaster().getDataBuffer()).getData()
	// So we emulate that same call here at the native level.
	jmethodID getRasterMeth = env->GetMethodID(buffImageClass, "getRaster", "()Ljava/awt/image/WritableRaster;");
	jobject rasterObj = env->CallObjectMethod(rv, getRasterMeth);
	jmethodID getDataBufferMeth = env->GetMethodID(env->FindClass("java/awt/image/Raster"), "getDataBuffer", "()Ljava/awt/image/DataBuffer;");
	jmethodID getDataMeth = env->GetMethodID(env->FindClass("java/awt/image/DataBufferInt"), "getData", "()[I");
	jarray javaPixelArrayObject = (jarray)env->CallObjectMethod(env->CallObjectMethod(rasterObj, getDataBufferMeth), getDataMeth);

	// Get the native memory pointer to the Java pixel array so we can copy the data into it
	void* rawJavaPixels = env->GetPrimitiveArrayCritical(javaPixelArrayObject, NULL);

	// Get the unmanaged memory pointer to the managed image data
	void* rawMgdPixels = mgdBitmapData->get_Scan0().ToPointer();

	// Copy the pixel data!
	memcpy(rawJavaPixels, rawMgdPixels, 4*imageWidth*imageHeight);

	// Release the hold on the java memory pointer for the pixels, be sure the changes are commited back to the JVM
	// memory if it was copied instead of pinned (that's what JNI_COMMIT does)
	env->ReleasePrimitiveArrayCritical(javaPixelArrayObject, rawJavaPixels, JNI_COMMIT);

	// Release the image data from the managed object we held
	mgdBitmap->UnlockBits(mgdBitmapData);

	// Return the Java image
	return rv;
}

System::String *JStringToMgdString(JNIEnv* env, jstring jstr)
{
	const jchar* jchars = env->GetStringChars(jstr, NULL);
	System::String* rv = new System::String(jchars, 0, (int) wcslen(jchars));
	env->ReleaseStringChars(jstr, jchars);
	return rv;
}
